@@ -18,7 +18,13 @@ import logging
 import math
 import sys
 
-__version__ = "0.12.1"
+__version__ = "0.14.0"
+
+
+log = logging.getLogger(__name__)
+log.addHandler(logging.NullHandler())
+
+_eps = sys.float_info.epsilon
 
 
 try:  # pragma: no cover
@@ -36,27 +42,36 @@ try:  # pragma: no cover
             label = "#" + str(count)
         return label
 
-    def bubbles(circles, labels=None, lim=None):
+    def _bubbles(circles, labels=None, lim=None):
         """Debugging function displays circles with matplotlib."""
-        _, ax = plt.subplots(figsize=(8.0, 8.0))
         if not labels:
-            labels = [get_default_label(i, c) for i, c in enumerate(circles)]
+            labels = range(len(circles))
+        _, ax = plt.subplots(figsize=(8.0, 8.0))
         for circle, label in zip(circles, labels):
-            x, y, r = circle.circle
+            x, y, r = circle
             ax.add_patch(plt.Circle((x, y), r, alpha=0.2, linewidth=2, fill=False))
             ax.text(x, y, label)
+        enclosure = enclose(circles)
+        n = len(circles)
+        if enclosure in circles:
+            n = n - 1
+        d = density(circles, enclosure)
+        title = "{} circles packed for density {:.4f}".format(n, d)
+        ax.set_title(title)
         if lim is None:
             lim = max(
-                max(
-                    abs(circle.circle.x) + circle.circle.r,
-                    abs(circle.circle.y) + circle.circle.r,
-                )
+                max(abs(circle.x) + circle.r, abs(circle.y) + circle.r,)
                 for circle in circles
             )
         plt.xlim(-lim, lim)
         plt.ylim(-lim, lim)
         plt.show()
 
+    def bubbles(circles, labels=None, lim=None):
+        if not labels:
+            labels = [get_default_label(i, c) for i, c in enumerate(circles)]
+        return _bubbles([c.circle for c in circles], labels, lim)
+
 
 except ImportError:  # pragma: no cover
     pass
@@ -127,12 +142,6 @@ class Circle:
         return self.circle.r
 
 
-log = logging.getLogger(__name__)
-log.addHandler(logging.NullHandler())
-
-_eps = sys.float_info.epsilon
-
-
 def distance(circle1, circle2):
     """Compute distance between 2 cirlces."""
     x1, y1, r1 = circle1
@@ -202,7 +211,7 @@ def get_placement_candidates(radius, c1, c2, margin):
         A pair of candidate cirlces where one or both value can be None.
 
     """
-    margin = 10.0 * sys.float_info.epsilon
+    margin = radius * _eps * 10.0
     ic1 = _Circle(c1.x, c1.y, c1.r + (radius + margin))
     ic2 = _Circle(c2.x, c2.y, c2.r + (radius + margin))
     i1, i2 = get_intersection(ic1, ic2)
@@ -217,36 +226,109 @@ def get_placement_candidates(radius, c1, c2, margin):
     return candidate1, candidate2
 
 
-def get_hole_degree(candidate, placed_circles, pc1, pc2):
+def get_hole_degree_radius_w(candidate, circles):
     """Calculate the hole degree of a candidate circle.
 
-    Note pc1 and pc2 must not be used in the evaluation of the minimum
-    distance.
-
     Args:
         candidate: candidate circle.
-        placed_circles: sequence of circles already placed.
-        pc1: first placed circle used to place the candidate.
-        pc2: second placed circle used to place the candidate.
+        circles: sequence of circles.
 
     Returns:
-        hole degree defined as (1 - d_min / r_i) where d_min is a minimum
-        disance between the candidate and the circles other than the one
-        used to place the candidate and r_i the radius of the candidate.
+        Squared euclidian distance of the candidate to the circles in argument.
+        Each component of the distance is weighted by the inverse of the radius
+        of the other circle to tilt the choice towards bigger circles.
 
     """
-    lsq = 0.0
-    for pc in placed_circles:
-        if pc1 is not None and pc1 == pc:
-            continue
-        if pc2 is not None and pc2 == pc:
-            continue
-        lsq += math.sqrt((candidate.y - pc.y) ** 2.0 + (candidate.x - pc.x) ** 2.0)
-    return -math.sqrt(lsq)
+    return sum(distance(candidate, c) * c.r for c in circles)
+
+
+def get_hole_degree_a1_0(candidate, circles):
+    """Calculate the hole degree of a candidate circle.
+
+    Args:
+        candidate: candidate circle.
+        circles: sequence of circles.
+
+    Returns:
+        minimum distance between the candidate and the circles in argument.
+
+    In the paper, the hole degree defined as (1 - d_min / r_i) where d_min is
+    a minimum disance between the candidate and the circles other than the one
+    used to place the candidate and r_i the radius of the candidate.
+
+    """
+    return min(distance(candidate, c) for c in circles)
+
+
+def get_hole_degree_density(candidate, circles):
+    """Calculate the hole degree of a candidate circle.
+
+    Args:
+        candidate: candidate circle.
+        circles: sequence of circles.
+
+    Returns:
+        One minus the density of the configuration. So the result should always
+        be between 0 and 1. See also density.
+
+    """
+    return 1.0 - density(circles + [candidate])
+
+
+def place_new_A1_0(radius, next_, const_placed_circles, get_hole_degree):
+    """Place a new circle.
+
+    Args:
+        radius: value to be added.
+        next_: next value to be added after radius.
+        const_placed_circles: circles.
+        get_hole_degree: objective function to maximize.
+
+    Returns:
+        New configuration.
+
+    """
+    placed_circles = const_placed_circles[:]
+    n_circles = len(placed_circles)
+    # If there are 2 or less, place circles on each side of (0, 0)
+    if n_circles <= 1:
+        x = radius if n_circles == 0 else -radius
+        circle = _Circle(x, float(0.0), radius)
+        placed_circles.append(circle)
+        return placed_circles
+    mhd = None
+    lead_candidate = None
+    for (c1, c2) in itertools.combinations(placed_circles, 2):
+        margin = radius * _eps * 10.0
+        # Placed circles other than the 2 circles used to find the
+        # candidate placement.
+        other_circles = [c for c in placed_circles if c not in (c1, c2)]
+        for cand in get_placement_candidates(radius, c1, c2, margin):
+            if cand is None:
+                continue
+            if not other_circles:
+                lead_candidate = cand
+                break
+            # If overlaps with any, skip this candidate.
+            if any(distance(c, cand) < 0.0 for c in other_circles):
+                continue
+            hd = get_hole_degree(cand, other_circles)
+            assert hd is not None, "hole degree should not be None!"
+            # If we were to use next_ we could use it here for look ahead.
+            if mhd is None or hd < mhd:
+                mhd = hd
+                lead_candidate = cand
+            if abs(mhd) < margin:
+                break
+    if lead_candidate is None:
+        # The radius is set to sqrt(value) in pack_A1_0
+        raise ValueError("cannot place circle for value " + str(radius ** 2))
+    placed_circles.append(lead_candidate)
+    return placed_circles
 
 
 def pack_A1_0(data):
-    """Pack circles whose radius is linked to the input data.
+    """Pack circles whose area is proportional to the input data.
 
     This algorithm is based on the Huang et al. heuristic.
 
@@ -257,40 +339,19 @@ def pack_A1_0(data):
         list of circlify.Output.
 
     """
+    min_max_ratio = min(data) / max(data)
+    if min_max_ratio < _eps:
+        log.warning(
+            "min to max ratio is too low at %f and it could cause algorithm stability issues. Try to remove insignificant data",
+            min_max_ratio,
+        )
     assert data == sorted(data, reverse=True), "data must be sorted (desc)"
     placed_circles = []
-    for value in data:
-        radius = math.sqrt(value)
-        n_circles = len(placed_circles)
-        # Place first 2 circles on each side of (0, 0)
-        if n_circles <= 1:
-            x = radius if n_circles == 0 else -radius
-            circle = _Circle(x, float(0.0), radius)
-            placed_circles.append(circle)
-            continue
-        mhd = None
-        lead_candidate = None
-        for (c1, c2) in itertools.combinations(placed_circles, 2):
-            margin = float(_eps)
-            for cand in get_placement_candidates(radius, c1, c2, margin):
-                if cand is not None:
-                    # Ignore candidates that overlap with any placed circle.
-                    other_placed_circles = [
-                        circ for circ in placed_circles if circ not in [c1, c2]
-                    ]
-                    if other_placed_circles:
-                        min_dist = min(
-                            distance(pc, cand) for pc in other_placed_circles
-                        )
-                        if min_dist < -_eps:
-                            continue
-                    hd = get_hole_degree(cand, placed_circles, c1, c2)
-                    if mhd is None or hd > mhd:
-                        mhd = hd
-                        lead_candidate = cand
-        if lead_candidate is None:
-            raise ValueError("cannot place circle for value %f", value)
-        placed_circles.append(lead_candidate)
+    radiuses = [math.sqrt(value) for value in data]
+    for radius, next_ in look_ahead(radiuses):
+        placed_circles = place_new_A1_0(
+            radius, next_, placed_circles, get_hole_degree_radius_w
+        )
     return placed_circles
 
 
@@ -314,7 +375,7 @@ def extendBasis(B, p):
                 and enclosesWeakAll(encloseBasis3(B[i], B[j], p), B)
             ):
                 return [B[i], B[j], p]
-    raise RuntimeError("If we get here then something is very wrong")
+    raise ValueError("If we get here then something is very wrong")
 
 
 def enclosesNot(a, b):
@@ -389,25 +450,6 @@ def encloseBasis3(a, b, c):
     return _Circle(x1 + xa + xb * r, y1 + ya + yb * r, r)
 
 
-def scale(circle, target, enclosure):
-    """Scale circle in enclosure to fit in the target circle.
-
-    Args:
-        circle: Circle to scale.
-        target: target Circle to scale to.
-        enclosure: allows one to specify the enclosure.
-
-    Returns:
-        scaled circle
-
-    """
-    r = target.r / enclosure.r
-    t_x, t_y = target.x, target.y
-    e_x, e_y = enclosure.x, enclosure.y
-    c_x, c_y, c_r = circle
-    return _Circle((c_x - e_x) * r + t_x, (c_y - e_y) * r + t_y, c_r * r)
-
-
 def enclose(circles):
     """Shamelessly adapted from d3js.
 
@@ -431,6 +473,53 @@ def enclose(circles):
     return e
 
 
+def scale(circle, target, enclosure):
+    """Scale circle in enclosure to fit in the target circle.
+
+    Args:
+        circle: Circle to scale.
+        target: target Circle to scale to.
+        enclosure: allows one to specify the enclosure.
+
+    Returns:
+        scaled circle
+
+    """
+    r = target.r / enclosure.r
+    t_x, t_y = target.x, target.y
+    e_x, e_y = enclosure.x, enclosure.y
+    c_x, c_y, c_r = circle
+    return _Circle((c_x - e_x) * r + t_x, (c_y - e_y) * r + t_y, c_r * r)
+
+
+def density(circles, enclosure=None):
+    """Computes the relative density of te packed circles.
+
+    Args:
+        circles: packed list of circles.
+
+    Return:
+        Compute the enclosure if not passed as argument and calculates the
+        relative area of the sum of the inner cirlces to the area of the
+        enclosure.
+
+    """
+    if not enclosure:
+        enclosure = enclose(circles)
+    return sum(c.r ** 2.0 for c in circles if c != enclosure) / enclosure.r ** 2.0
+
+
+def look_ahead(iterable, n_elems=1):
+    """Fetch look ahead elements of data
+
+    From https://stackoverflow.com/questions/4197805/python-for-loop-look-ahead
+
+    """
+    items, nexts = itertools.tee(iterable, 2)
+    nexts = itertools.islice(nexts, n_elems, None)
+    return itertools.zip_longest(items, nexts)
+
+
 def _handle(data, level, fields=None):
     """Converts possibly heterogeneous list of float or dict in list of Output.
 
@@ -447,11 +536,18 @@ def _handle(data, level, fields=None):
         if isinstance(datum, dict):
             value = datum[datum_field]
             elements.append(Circle(r=value + 0, level=level, ex=datum))
-        else:
-            try:
-                elements.append(Circle(r=datum + 0, level=level, ex={"datum": datum}))
-            except TypeError:  # if it fails, assume dict.
-                raise TypeError("dict or numeric value expected")
+            continue
+        if datum <= 0.0:
+            raise ValueError("input data must be positive. Found " + str(datum))
+        if datum <= _eps:
+            log.warning(
+                "input data %f is small and could cause stability issues. Can you scale the data set up or drop insignificant elements?",
+                datum,
+            )
+        try:
+            elements.append(Circle(r=datum + 0, level=level, ex={"datum": datum}))
+        except TypeError:  # if it fails, assume dict.
+            raise TypeError("dict or numeric value expected")
     return sorted(elements, reverse=True)
 
 
@@ -508,7 +604,7 @@ def circlify(
     id_field="id",
     children_field="children",
 ):
-    """Pack and enclose circles whose radius is linked to the input data.
+    """Pack and enclose circles.
 
     Args:
         data: sorted (descending) array of values.
@@ -522,8 +618,9 @@ def circlify(
             element is a dict.
 
     Returns:
-        list of circlify.Circle sorted by ascending level (root to leaf) and
-        descending value (biggest circles first).
+        list of circlify.Circle whose *area* is proportional to the
+        corresponding input value.  The list is sorted by ascending level
+        (root to leaf) and descending value (biggest circles first).
 
     """
     fields = FieldNames(id=id_field, datum=datum_field, children=children_field)
